@@ -145,8 +145,7 @@ static errcode_t implied_cluster_alloc(ext2_filsys fs, ext2_ino_t ino,
 	blk64_t	base_block, pblock = 0;
 	int i;
 
-	if (!EXT2_HAS_RO_COMPAT_FEATURE(fs->super,
-					EXT4_FEATURE_RO_COMPAT_BIGALLOC))
+	if (!ext2fs_has_feature_bigalloc(fs->super))
 		return 0;
 
 	base_block = lblk & ~EXT2FS_CLUSTER_MASK(fs);
@@ -183,8 +182,7 @@ errcode_t ext2fs_map_cluster_block(ext2_filsys fs, ext2_ino_t ino,
 
 	/* Need bigalloc and extents to be enabled */
 	*pblk = 0;
-	if (!EXT2_HAS_RO_COMPAT_FEATURE(fs->super,
-					EXT4_FEATURE_RO_COMPAT_BIGALLOC) ||
+	if (!ext2fs_has_feature_bigalloc(fs->super) ||
 	    !(inode->i_flags & EXT4_EXTENTS_FL))
 		return 0;
 
@@ -209,15 +207,19 @@ static errcode_t extent_bmap(ext2_filsys fs, ext2_ino_t ino,
 			     int *ret_flags, int *blocks_alloc,
 			     blk64_t *phys_blk)
 {
+	struct blk_alloc_ctx	alloc_ctx;
 	struct ext2fs_extent	extent;
 	unsigned int		offset;
 	errcode_t		retval = 0;
 	blk64_t			blk64 = 0;
 	int			alloc = 0;
+	int			set_flags;
+
+	set_flags = bmap_flags & BMAP_UNINIT ? EXT2_EXTENT_SET_BMAP_UNINIT : 0;
 
 	if (bmap_flags & BMAP_SET) {
 		retval = ext2fs_extent_set_bmap(handle, block,
-						*phys_blk, 0);
+						*phys_blk, set_flags);
 		return retval;
 	}
 	retval = ext2fs_extent_goto(handle, block);
@@ -244,9 +246,13 @@ got_block:
 		retval = extent_bmap(fs, ino, inode, handle, block_buf,
 				     0, block-1, 0, blocks_alloc, &blk64);
 		if (retval)
-			blk64 = 0;
-		retval = ext2fs_alloc_block2(fs, blk64, block_buf,
-					     &blk64);
+			blk64 = ext2fs_find_inode_goal(fs, ino, inode, block);
+		alloc_ctx.ino = ino;
+		alloc_ctx.inode = inode;
+		alloc_ctx.lblk = extent.e_lblk;
+		alloc_ctx.flags = BLOCK_ALLOC_DATA;
+		retval = ext2fs_alloc_block3(fs, blk64, block_buf, &blk64,
+					     &alloc_ctx);
 		if (retval)
 			return retval;
 		blk64 &= ~EXT2FS_CLUSTER_MASK(fs);
@@ -254,7 +260,7 @@ got_block:
 		alloc++;
 	set_extent:
 		retval = ext2fs_extent_set_bmap(handle, block,
-						blk64, 0);
+						blk64, set_flags);
 		if (retval) {
 			ext2fs_block_alloc_stats2(fs, blk64, -1);
 			return retval;
@@ -299,9 +305,16 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 	ext2_extent_handle_t handle = 0;
 	blk_t addr_per_block;
 	blk_t	b, blk32;
+	blk64_t b64;
 	char	*buf = 0;
 	errcode_t	retval = 0;
 	int		blocks_alloc = 0, inode_dirty = 0;
+	struct blk_alloc_ctx alloc_ctx = {
+		.ino	= ino,
+		.inode	= inode,
+		.lblk	= 0,
+		.flags	= BLOCK_ALLOC_DATA,
+	};
 
 	if (!(bmap_flags & BMAP_SET))
 		*phys_blk = 0;
@@ -321,6 +334,13 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 	if (ext2fs_file_block_offset_too_big(fs, inode, block))
 		return EXT2_ET_FILE_TOO_BIG;
 
+	/*
+	 * If an inode has inline data, that means that it doesn't have
+	 * any blocks and we shouldn't map any blocks for it.
+	 */
+	if (inode->i_flags & EXT4_INLINE_DATA_FL)
+		return EXT2_ET_INLINE_DATA_NO_BLOCK;
+
 	if (!block_buf) {
 		retval = ext2fs_get_array(2, fs->blocksize, &buf);
 		if (retval)
@@ -347,10 +367,14 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 		}
 
 		*phys_blk = inode_bmap(inode, block);
-		b = block ? inode_bmap(inode, block-1) : 0;
+		b = block ? inode_bmap(inode, block - 1) :
+			    ext2fs_find_inode_goal(fs, ino, inode, block);
 
 		if ((*phys_blk == 0) && (bmap_flags & BMAP_ALLOC)) {
-			retval = ext2fs_alloc_block(fs, b, block_buf, &b);
+			b64 = b;
+			retval = ext2fs_alloc_block3(fs, b64, block_buf, &b64,
+						     &alloc_ctx);
+			b = b64;
 			if (retval)
 				goto done;
 			inode_bmap(inode, block) = b;
@@ -373,7 +397,10 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 			}
 
 			b = inode_bmap(inode, EXT2_IND_BLOCK-1);
- 			retval = ext2fs_alloc_block(fs, b, block_buf, &b);
+			b64 = b;
+			retval = ext2fs_alloc_block3(fs, b64, block_buf, &b64,
+						     &alloc_ctx);
+			b = b64;
 			if (retval)
 				goto done;
 			inode_bmap(inode, EXT2_IND_BLOCK) = b;
@@ -398,7 +425,10 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 			}
 
 			b = inode_bmap(inode, EXT2_IND_BLOCK);
- 			retval = ext2fs_alloc_block(fs, b, block_buf, &b);
+			b64 = b;
+			retval = ext2fs_alloc_block3(fs, b64, block_buf, &b64,
+						     &alloc_ctx);
+			b = b64;
 			if (retval)
 				goto done;
 			inode_bmap(inode, EXT2_DIND_BLOCK) = b;
@@ -422,7 +452,10 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 		}
 
 		b = inode_bmap(inode, EXT2_DIND_BLOCK);
-		retval = ext2fs_alloc_block(fs, b, block_buf, &b);
+		b64 = b;
+		retval = ext2fs_alloc_block3(fs, b64, block_buf, &b64,
+					     &alloc_ctx);
+		b = b64;
 		if (retval)
 			goto done;
 		inode_bmap(inode, EXT2_TIND_BLOCK) = b;
@@ -433,6 +466,8 @@ errcode_t ext2fs_bmap2(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode,
 	if (retval == 0)
 		*phys_blk = blk32;
 done:
+	if (*phys_blk && retval == 0 && (bmap_flags & BMAP_ZERO))
+		retval = ext2fs_zero_blocks2(fs, *phys_blk, 1, NULL, NULL);
 	if (buf)
 		ext2fs_free_mem(&buf);
 	if (handle)
