@@ -115,7 +115,7 @@ static char *undo_file;
 static void usage(void)
 {
 	fprintf(stderr,
-		_("Usage: %s [-f] [-h] [-n] [-v] <transaction file> <filesystem>\n"), prg_name);
+		_("Usage: %s [-f] [-h] [-n] [-o offset] [-v] [-z undo_file] <transaction file> <filesystem>\n"), prg_name);
 	exit(1);
 }
 
@@ -208,8 +208,7 @@ static int key_compare(const void *a, const void *b)
 
 	ka = a;
 	kb = b;
-	return ext2fs_le64_to_cpu(ka->fsblk) -
-	       ext2fs_le64_to_cpu(kb->fsblk);
+	return ka->fsblk - kb->fsblk;
 }
 
 static int e2undo_setup_tdb(const char *name, io_manager *io_ptr)
